#include <stdlib.h>
#include <string.h>
#include <cuda.h>
#include <stdio.h>

#include "StcCell_Cuda_Interface.h"
#include "StcCell_Container.h"

#ifndef CMCF_INTERACTION
#include "CellML_MahajanCM.h"
#include "CellML_PanditRatCM.h"
#include "Cell_MacCannellCF.h"
#include "Cell_LuoRudyCM.h"
#endif

#include "Cell_MacCannellINaCF.h"
#include "Cell_PassiveCF.h"
#include "Cell_PassiveIVCF.h"
#include "Cell_LuoRudy1CM.h"

// variables inside 
// currently 1 -3D dimension using general diffusion scheme
// static can be added


static int iCudaBlockSizeLocal = iCudaBlockSize;
//static int iCellDim;
//static int myo_xdim, myo_ydim, myo_zdim, myo_total;

// container
static StcCellContainer *pCtn;
static int hasCudaCells = 0;

// diffusion in neighbor index
static StcCellDiffusion *pDiff;
static int hasDiffusion = 0;

static TypeCellML dt; // time step

//static unsigned int 
size_t iByteSizeState, iByteSizeRate, iByteSizeConstant, iByteSizeAlgebraic;


/* interface */

int cuda_set_DeviceId(int iDev)
{
	cudaError_t iErr;
	iErr = cudaSetDevice(iDev);
	if (iErr == cudaSuccess) return 1;
	else return 0;
}

void cuda_meminfo(size_t *available, size_t *total)
{
	cudaMemGetInfo(available, total);
}


void cuda_set_BlockSize(int iSize)
{
	iCudaBlockSizeLocal = iSize;
}

void cuda_set_dt(TypeCellML val) {
	dt = val;
}

// input - integer array, size of 3
// 0: total number of cells, 1: number of cell types, 2: mem alloc
void cuda_get_info_ctn(int *pInfo)
{
	pInfo[0] = pCtn->iNCell;
	pInfo[1] = pCtn->iNStcCell;
	pInfo[2] = pCtn->iMemFlag_temp;
}

void cuda_get_info_cell(int iCell, int *pInfo)
{
	StcCell *pCell;
	pCell = pCtn->pCell[iCell];
	pInfo[0] = pCell->N_CELL;
	pInfo[1] = pCell->N_STATE;
	pInfo[2] = pCell->N_CONST;
}

// input : size of 3, integer array
void cuda_get_info_diffusion(int *pInfo)
{
	pInfo[0] = pDiff->iNCell;
	pInfo[1] = pDiff->iNCoupling;
	pInfo[2] = pDiff->isAllocated;
}

// allocate container and diffusion
void cuda_alloc_ctn(void)
{
	pCtn = ctn_prototype();
}

void cuda_free_ctn(void)
{
	ctn_freeStcCellContainer(pCtn);
}


void cuda_alloc_diffusion(int iMaxCoupling)
{
	pDiff = diff_prototype_StcCellDiffusion(pCtn, iMaxCoupling);
}

void cuda_free_diffusion(void)
{
	diff_free_StcCellDiffusion(pDiff);
}


// add cells into container

void cuda_add_cell(enum CellType iType, int iNCell)
{
	StcCell *pStc;

	pStc = NULL;
	switch(iType) {
#ifndef CMCF_INTERACTION
		case CELL_MAHAJAN_CM:
			pStc = prototype_Cell_Mahajan();
			break;
		case CELL_PANDIT_RATCM:
			pStc = prototype_Cell_PanditRatCM();
			break;
		case CELL_MACCANNELL_CF:
			pStc = prototype_Cell_MacCannellCF();
			break;
		case CELL_LUORUDY_CM:
			pStc = prototype_Cell_LuoRudyCM();
			break;
#endif
		case CELL_MACCANNELLINA_CF:
			pStc = prototype_Cell_MacCannellINaCF();
			break;
		case CELL_PASSIVE_CF:
			pStc = prototype_Cell_PassiveCF();
			break;
		case CELL_LUORUDY1_CM:
			pStc = prototype_Cell_LuoRudy1CM();
			break;
		case CELL_PASSIVEIV_CF:
			pStc = prototype_Cell_PassiveIVCF();
			break;
	}
	if (pStc != NULL) {
		pStc->iCellType = iType;
		pStc->N_CELL = iNCell;
		cell_alloc_StcCellOnCuda(pStc);
		(*(pStc->fpInitCellPar))(pStc);
		ctn_addCell(pCtn, pStc);
	}
}


void cuda_set_CellLoc(int iCell, int *pVal)
{
	StcCell  *pCell;
	pCell = pCtn->pCell[iCell];
	cudaMemcpy(pCell->d_CellLoc, pVal, sizeof(int)*pCell->N_CELL, cudaMemcpyHostToDevice);
}


void cuda_get_CellLoc(int iCell, int *pVal)
{
	StcCell  *pCell;
	pCell = pCtn->pCell[iCell];
	cudaMemcpy(pVal, pCell->d_CellLoc, sizeof(int)*pCell->N_CELL, cudaMemcpyDeviceToHost);
}


// interface to set and retrieve values
void cuda_set_CellValue(int iCell, TypeCellML *pVal, int iPar, CellMLVarType iVarType)
{
	StcCell *pCell;
	pCell = pCtn->pCell[iCell];
	cell_set_CellValue(pCell, pVal, iPar, iVarType);
}

void cuda_get_CellValue(int iCell, TypeCellML *pVal, int iPar, CellMLVarType iVarType)
{
	StcCell *pCell;
	pCell = pCtn->pCell[iCell];
	cell_get_CellValue(pCell, pVal, iPar, iVarType);
}

void cuda_set_Cell_State(int iCell, TypeCellML *pState)
{
	StcCell *pCell;
	pCell = pCtn->pCell[iCell];
	cell_copy_VarToCuda(pCell, pState, State);
}


void cuda_set_Cell_Const(int iCell, TypeCellML *pConst)
{
	StcCell *pCell;
	pCell = pCtn->pCell[iCell];
	cell_copy_VarToCuda(pCell, pConst, Const);
}


void cuda_get_Cell_State(int iCell, TypeCellML *pState)
{
	StcCell *pCell;
	pCell = pCtn->pCell[iCell];
	cell_copy_VarFromCuda(pCell, pState, State);
}


void cuda_get_Cell_Const(int iCell, TypeCellML *pConst)
{
	StcCell *pCell;
	pCell = pCtn->pCell[iCell];
	cell_copy_VarFromCuda(pCell, pConst, Const);
}


void cuda_get_ValueSorted(TypeCellML *pVal, int *iIndCell, enum CellMLVarType iVarType)
{
	ctn_getValueSorted_device(pCtn, iIndCell, iVarType);
	ctn_getCtnValueToHost(pCtn, pVal);
}	

void cuda_get_Vm(TypeCellML *pVal)
{
	ctn_getVm(pCtn, pVal);
}

void cuda_set_Vm(TypeCellML *pVal)
{
	ctn_setVm(pCtn, pVal);
}

void cuda_set_stim(TypeCellML *pVal)
{
	ctn_setTemp(pCtn, pVal);
	ctn_setStim(pCtn);
}



// diffusion interface
void cuda_set_DiffParm(int *pInd, TypeCellML *pCoeff)
{
	diff_set_DiffParm(pDiff, pInd, pCoeff);
}

void cuda_calc_Diffusion(void)
{
	diff_calc_Diffusion(pDiff);
}

void cuda_simulate_SingleStep(void){
	ctn_iterate(pCtn, dt);
	cudaDeviceSynchronize();
	// cuda_calcDiffusion();
}


void cuda_simulate_Nsteps(int iStep) {
	int i;
	for (i=0;i<iStep;i++) {
		cuda_simulate_SingleStep();
		cuda_calc_Diffusion();
	}
	//cudaDeviceSynchronize();
}

void cuda_simulate_NstepsNoDiffusion(int iStep) {
	int i;
	for (i=0;i<iStep;i++) {
		cuda_simulate_SingleStep();
		// cuda_calcDiffusion();
	}
	//cudaDeviceSynchronize();
}

